from os import PathLike
from pathlib import Path
from typing import Union
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile


def zip_dir(zip_name: str, source_dir: Union[str, PathLike]):
    src_path = Path(source_dir).expanduser().resolve(strict=True)
    with ZipFile(zip_name, "w", ZIP_DEFLATED) as zf:
        for file in src_path.rglob("*"):
            zf.write(file, file.relative_to(src_path))



import os

from notion_client.helpers import get_id

from notion2md.exceptions import MissingTargetIDError


class Config(object):
    __slots__ = (
        "file_name",
        "target_id",
        "output_path",
        "tmp_path",
        "download",
        "unzipped",
        "path_name",
    )

    def __init__(
        self,
        block_id: str = None,
        block_url: str = None,
        output_filename: str = None,
        output_path: str = None,
        download: bool = False,
        unzipped: bool = False,
    ):
        if block_url:
            self.target_id = get_id(block_url)
        elif block_id:
            self.target_id = block_id
        else:
            raise MissingTargetIDError

        if output_filename:
            self.file_name = output_filename
        else:
            self.file_name = self.target_id

        if output_path:
            self.path_name = output_path
            self.output_path = os.path.abspath(output_path)

        else:
            self.path_name = "notion2md-output"
            self.output_path = os.path.join(os.getcwd(), "notion2md-output")

        if download:
            self.download = True
        else:
            self.download = False

        if unzipped:
            self.unzipped = True
            self.tmp_path = self.output_path
        else:
            self.unzipped = False
            self.tmp_path = os.path.join(os.getcwd(), "tmp")



import os
import shutil

from notion2md.config import Config
from notion2md.convertor.block import BlockConvertor
from notion2md.notion_api import NotionClient
from notion2md.util import zip_dir


class Exporter:
    def __init__(
        self,
        block_id: str = None,
        block_url: str = None,
        output_filename: str = None,
        output_path: str = None,
        download: bool = False,
        unzipped: bool = False,
    ):
        self._config = Config(
            block_id=block_id,
            block_url=block_url,
            output_filename=output_filename,
            output_path=output_path,
            download=download,
            unzipped=unzipped,
        )
        self._client = NotionClient()
        self._io = None
        self._block_convertor = None

    @property
    def block_convertor(self):
        if not self._block_convertor:
            self._block_convertor = BlockConvertor(
                self._config, self._client, self._io
            )
        return self._block_convertor

    @property
    def config(self):
        return self._config

    @property
    def io(self):
        return self._io

    @io.setter
    def io(self, io):
        self._io = io

    def create_directories(self):
        if not os.path.exists(self._config.tmp_path):
            os.makedirs(self._config.tmp_path)
        if not os.path.exists(self._config.output_path):
            os.mkdir(self._config.output_path)

    def get_blocks(self):
        return self._client.get_children(self._config.target_id)

    def make_zip(self):
        zip_dir(
            os.path.join(self._config.output_path, self._config.file_name)
            + ".zip",
            self._config.tmp_path,
        )
        shutil.rmtree(self._config.tmp_path)

    def export(self):
        pass


class MarkdownExporter(Exporter):
    def export(self):
        self.create_directories()
        with open(
            os.path.join(
                self._config.tmp_path, self._config.file_name + ".md"
            ),
            "w",
            encoding="utf-8",
        ) as output:
            output.write(self.block_convertor.convert(self.get_blocks()))
        if not self._config.unzipped:
            self.make_zip()


class StringExporter(Exporter):
    def export(self):
        return self.block_convertor.to_string(self.get_blocks())


class CLIExporter(Exporter):
    def export(self, blocks):
        with open(
            os.path.join(
                self._config.tmp_path, self._config.file_name + ".md"
            ),
            "w",
            encoding="utf-8",
        ) as output:
            output.write(self.block_convertor.convert(blocks))



import os

from notion_client import Client

from notion2md.exceptions import MissingTokenError


def singleton(cls):
    instance = {}

    def get_instance():
        if cls not in instance:
            instance[cls] = cls()
        return instance[cls]

    return get_instance


@singleton
class NotionClient:
    def __init__(self):
        token = self._get_env_variable()
        self._client = Client(auth=token)

    def _get_env_variable(self):
        try:
            return os.environ["NOTION_TOKEN"]
        except Exception:
            raise MissingTokenError() from None

    def get_children(self, parent_id):
        # Most pages are small
        results = []
        start_cursor = None
        # Avoid infinite loops
        for _ in range(100):
            resp = self._client.blocks.children.list(
                parent_id, start_cursor=start_cursor, page_size=100
            )
            results.extend(resp["results"])
            start_cursor = resp["next_cursor"] if resp["has_more"] else None
            if start_cursor is None:
                return results
        raise Exception(
            "Can't parse notion page of > 10,000 children! (e.g. blocks)"
        )



import concurrent.futures
import hashlib
import os
import urllib.request as request
import uuid

from urllib.parse import urlparse

from cleo.io.io import IO

from notion2md.config import Config
from notion2md.console.formatter import error
from notion2md.console.formatter import status
from notion2md.console.formatter import success
from notion2md.notion_api import NotionClient

from .richtext import richtext_convertor


class BlockConvertor:
    def __init__(self, config: Config, client: NotionClient, io: IO = None):
        self._config = config
        self._client = client
        self._io = io
        self._continued_numbered_list = False
        self._numbered_list_number = 1

    def convert(self, blocks: dict) -> str:
        outcome_blocks: str = ""
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(self.convert_block, blocks)
            outcome_blocks = "".join([result for result in results])
        return outcome_blocks

    def convert_block(self, block: dict, depth=0) -> str:
        outcome_block: str = ""
        block_type = block["type"]
        # Handle the case where the block is a list item
        if block_type == "numbered_list_item":
            if self._continued_numbered_list:
                self._numbered_list_number += 1
            else:
                self._continued_numbered_list = True
                self._numbered_list_number = 1
        else:
            self._continued_numbered_list = False
        # Special Case: Block is blank
        if (
            block_type == "paragraph"
            and not block["has_children"]
            and not block[block_type]["rich_text"]
        ):
            return blank() + "\n\n"
        # Normal Case
        try:
            if block_type in BLOCK_TYPES:
                outcome_block = (
                    BLOCK_TYPES[block_type](
                        self.collect_info(block[block_type])
                    )
                    + "\n\n"
                )
            else:
                outcome_block = f"[//]: # ({block_type} is not supported)\n\n"
            # Convert child block
            if block["has_children"]:
                # create child page
                if block_type == "child_page":
                    # call make_child_function
                    pass
                # create table block
                elif block_type == "table":
                    depth += 1
                    child_blocks = self._client.get_children(block["id"])
                    outcome_block = self.create_table(cell_blocks=child_blocks)
                # create indent block
                else:
                    depth += 1
                    child_blocks = self._client.get_children(block["id"])
                    for block in child_blocks:
                        outcome_block += "\t" * depth + self.convert_block(
                            block, depth
                        )
        except Exception as e:
            if self._io:
                self._io.write_line(
                    error(f"{e}: Error occured block_type:{block_type}")
                )
        return outcome_block

    def create_table(self, cell_blocks: dict):
        table_list = []
        for cell_block in cell_blocks:
            cell_block_type = cell_block["type"]
            table_list.append(
                BLOCK_TYPES[cell_block_type](
                    self.collect_info(cell_block[cell_block_type])
                )
            )
        # convert to markdown table
        for index, value in enumerate(table_list):
            if index == 0:
                table = " | " + " | ".join(value) + " | " + "\n"
                table += (
                    " | " + " | ".join(["----"] * len(value)) + " | " + "\n"
                )
                continue
            table += " | " + " | ".join(value) + " | " + "\n"
        table += "\n"
        return table

    def collect_info(self, payload: dict) -> dict:
        info = dict()
        if "rich_text" in payload:
            info["text"] = richtext_convertor(payload["rich_text"])
        if "icon" in payload:
            info["icon"] = payload["icon"]["emoji"]
        if "checked" in payload:
            info["checked"] = payload["checked"]
        if "expression" in payload:
            info["text"] = payload["expression"]
        if "url" in payload:
            info["url"] = payload["url"]
        if "caption" in payload:
            info["caption"] = richtext_convertor(payload["caption"])
        if "external" in payload:
            info["url"] = payload["external"]["url"]
            name, file_path = self.download_file(info["url"])
            info["file_name"] = name
            info["file_path"] = file_path
        if "language" in payload:
            info["language"] = payload["language"]
        # interal url
        if "file" in payload:
            info["url"] = payload["file"]["url"]
            name, file_path = self.download_file(info["url"])
            info["file_name"] = name
            info["file_path"] = file_path
        # table cells
        if "cells" in payload:
            info["cells"] = payload["cells"]
        info["number"] = self._numbered_list_number
        return info

    def download_file(self, url: str) -> str:
        file_name = os.path.basename(urlparse(url).path)
        if self._config.download:
            if file_name:
                name, extension = os.path.splitext(file_name)

                if not extension:
                    return file_name, url

                url_hash = hashlib.blake2s(
                    urlparse(url).path.encode()
                ).hexdigest()[:8]
                downloaded_file_name = f"{url_hash}_{file_name}"

                fullpath = os.path.join(
                    self._config.tmp_path, downloaded_file_name
                )

                if self._io:
                    self._io.write_line(status("Downloading", f"{file_name}"))
                    request.urlretrieve(url, fullpath)
                    self._io.write_line(
                        success(
                            "Downloaded",
                            f'"{file_name}" -> "{downloaded_file_name}"',
                        )
                    )
                else:
                    request.urlretrieve(url, fullpath)
                return name, downloaded_file_name
            else:
                if self._io:
                    self._io.write_line(error(f"invalid {url}"))
        else:
            return file_name, url

    def to_string(self, blocks: dict) -> str:
        return self.convert(blocks)


# Converting Methods
def paragraph(info: dict) -> str:
    return info["text"]


def heading_1(info: dict) -> str:
    return f"# {info['text']}"


def heading_2(info: dict) -> str:
    return f"## {info['text']}"


def heading_3(info: dict) -> str:
    return f"### {info['text']}"


def callout(info: dict) -> str:
    return f"{info['icon']} {info['text']}"


def quote(info: dict) -> str:
    return f"> {info['text']}"


# toggle item will be changed as bulleted list item
def bulleted_list_item(info: dict) -> str:
    return f"- {info['text']}"


# numbering is not supported
def numbered_list_item(info: dict) -> str:
    """
    input: item:dict = {"number":int, "text":str}
    """
    return f"{info['number']}. {info['text']}"


def to_do(info: dict) -> str:
    """
    input: item:dict = {"checked":bool, "test":str}
    """
    return f"- {'[x]' if info['checked'] else '[ ]'} {info['text']}"


# not yet supported
# child_database will be changed as child page
# def child_page(info:dict) -> str:
#     """
#     input: item:dict = {"id":str,"text":str}
#     """
#     #make_page(info['id'])
#     text = info['text']
#     return f'[{text}]({text})'


def code(info: dict) -> str:
    """
    input: item:dict = {"language":str,"text":str}
    """
    return f"\n```{info['language']}\n{info['text']}\n```"


def embed(info: dict) -> str:
    """
    input: item:dict ={"url":str,"text":str}
    """
    return f"[{info['url']}]({info['url']})"


def image(info: dict) -> str:
    """
    input: item:dict ={"url":str,"text":str,"caption":str}
    """
    # name,file_path = downloader(info['url'])

    if info["caption"]:
        return (
            f"![{info['file_name']}]({info['file_path']})\n\n{info['caption']}"
        )
    else:
        return f"![{info['file_name']}]({info['file_path']})"


def file(info: dict) -> str:
    # name,file_path = downloader(info['url'])
    return f"[{info['file_name']}]({info['file_path']})"


def bookmark(info: dict) -> str:
    """
    input: item:dict ={"url":str,"text":str,"caption":str}
    """
    if info["caption"]:
        return f"[{info['url']}]({info['url']})\n\n{info['caption']}"
    else:
        return f"[{info['url']}]({info['url']})"


def equation(info: dict) -> str:
    return f"$$ {info['text']} $$"


def divider(info: dict) -> str:
    return "---"


def blank() -> str:
    return "<br/>"


def table_row(info: list) -> list:
    """
    input: item:list = [[richtext],....]
    """
    column_list = []
    for column in info["cells"]:
        column_list.append(richtext_convertor(column))
    return column_list


# Since Synced Block has only child blocks, not name, it will return blank
def synced_block(info: list) -> str:
    return "[//]: # (Synced Block)"


# Block type map
BLOCK_TYPES = {
    "paragraph": paragraph,
    "heading_1": heading_1,
    "heading_2": heading_2,
    "heading_3": heading_3,
    "callout": callout,
    "toggle": bulleted_list_item,
    "quote": quote,
    "bulleted_list_item": bulleted_list_item,
    "numbered_list_item": numbered_list_item,
    "to_do": to_do,
    # "child_page": child_page,
    "code": code,
    "embed": embed,
    "image": image,
    "bookmark": bookmark,
    "equation": equation,
    "divider": divider,
    "file": file,
    "table_row": table_row,
    "synced_block": synced_block,
}



class MissingTokenError(Exception):
    """Exception for missing notion token error.

    Notion2Md requires Notion Integration token key to export Notion page to Markdown.
    The token should be saved as envronmental variable. If it is not, this exception
    is raised.
    """

    def __init__(self) -> None:
        super().__init__("Envrionment Variable 'NOTION_TOKEN' is not found")


class MissingTargetIDError(Exception):
    """Exception for missing Notion page's id error.

    Notion2Md requires url or id of Notion's page. This exception is raised if there
    is no target id from user input.
    """

    def __init__(self) -> None:
        super().__init__("Notion Page's id or url is not given as argument")



def error(msg):
    return f"\n<error>{'ERROR'+' ':>12}</error>{msg}\n"


def status(status, msg):
    return f"<status>{status+' ':>12}</status>{msg}"


def success(status, msg):
    return f"<success>{status+' ':>12}</success>{msg}"
